@@ -6,8 +6,9 @@
 #define ALIGNER_23MM_H_
 
 #include <utility>
-#include <vector>
+
 #include "aligner.h"
+#include "ds.h"
 #include "hit.h"
 #include "range_source.h"
 #include "row_chaser.h"
@@ -20,11 +21,11 @@
 class Unpaired23mmAlignerV1Factory : public AlignerFactory {
 	typedef RangeSourceDriver<EbwtRangeSource> TRangeSrcDr;
 	typedef CostAwareRangeSourceDriver<EbwtRangeSource> TCostAwareRangeSrcDr;
-	typedef std::vector<TRangeSrcDr*> TRangeSrcDrPtrVec;
+	typedef EList<TRangeSrcDr*> TRangeSrcDrPtrVec;
 public:
 	Unpaired23mmAlignerV1Factory(
-			Ebwt<String<Dna> >& ebwtFw,
-			Ebwt<String<Dna> >* ebwtBw,
+			Ebwt& ebwtFw,
+			Ebwt* ebwtBw,
 			bool two,
 			bool doFw,
 			bool doRc,
@@ -35,7 +36,7 @@ public:
 			uint32_t cacheLimit,
 			ChunkPool *pool,
 			BitPairReference* refs,
-			vector<String<Dna5> >& os,
+			EList<BTRefString >& os,
 			bool maqPenalty,
 			bool qualOrder,
 			bool strandFix,
@@ -60,8 +61,7 @@ public:
 			strandFix_(strandFix),
 			rangeMode_(rangeMode),
 			verbose_(verbose),
-			quiet_(quiet),
-			seed_(seed)
+			quiet_(quiet)
 	{
 		assert(ebwtFw.isInMemory());
 		assert(ebwtBw != NULL);
@@ -74,21 +74,21 @@ public:
 	virtual Aligner* create() const {
 
 		HitSinkPerThread* sinkPt = sinkPtFactory_.create();
-		EbwtSearchParams<String<Dna> >* params =
-			new EbwtSearchParams<String<Dna> >(*sinkPt, os_);
+		EbwtSearchParams* params =
+			new EbwtSearchParams(*sinkPt, os_);
 
 		const bool seeded = false;
 
 		EbwtRangeSource *rFw_Bw = new EbwtRangeSource(
-			 ebwtBw_, true, 0xffffffff, true, verbose_, quiet_, 0, seeded, maqPenalty_, qualOrder_);
+			 ebwtBw_, true, OFF_MASK, true, verbose_, quiet_, 0, seeded, maqPenalty_, qualOrder_);
 		EbwtRangeSource *rFw_Fw = new EbwtRangeSource(
-			&ebwtFw_, true, 0xffffffff, false, verbose_, quiet_, 0, seeded, maqPenalty_, qualOrder_);
+			&ebwtFw_, true, OFF_MASK, false, verbose_, quiet_, 0, seeded, maqPenalty_, qualOrder_);
 		EbwtRangeSource *rFw_BwHalf = new EbwtRangeSource(
-			 ebwtBw_, true, 0xffffffff, false, verbose_, quiet_, 2,  seeded, maqPenalty_, qualOrder_);
+			 ebwtBw_, true, OFF_MASK, false, verbose_, quiet_, 2,  seeded, maqPenalty_, qualOrder_);
 		EbwtRangeSource *rFw_FwHalf = NULL;
 		if(!two_) {
 			rFw_FwHalf = new EbwtRangeSource(
-				&ebwtFw_, true, 0xffffffff, false, verbose_, quiet_, 3,  seeded, maqPenalty_, qualOrder_);
+				&ebwtFw_, true, OFF_MASK, false, verbose_, quiet_, 3,  seeded, maqPenalty_, qualOrder_);
 		}
 
 		// Driver wrapper for rFw_Bw
@@ -146,15 +146,15 @@ public:
 		}
 
 		EbwtRangeSource *rRc_Fw = new EbwtRangeSource(
-			&ebwtFw_, false, 0xffffffff, true,  verbose_, quiet_, 0, seeded, maqPenalty_, qualOrder_);
+			&ebwtFw_, false, OFF_MASK, true,  verbose_, quiet_, 0, seeded, maqPenalty_, qualOrder_);
 		EbwtRangeSource *rRc_Bw = new EbwtRangeSource(
-			 ebwtBw_, false, 0xffffffff, false, verbose_, quiet_, 0, seeded, maqPenalty_, qualOrder_);
+			 ebwtBw_, false, OFF_MASK, false, verbose_, quiet_, 0, seeded, maqPenalty_, qualOrder_);
 		EbwtRangeSource *rRc_FwHalf = new EbwtRangeSource(
-			&ebwtFw_, false, 0xffffffff, false, verbose_, quiet_, 2,  seeded, maqPenalty_, qualOrder_);
+			&ebwtFw_, false, OFF_MASK, false, verbose_, quiet_, 2,  seeded, maqPenalty_, qualOrder_);
 		EbwtRangeSource *rRc_BwHalf = NULL;
 		if(!two_) {
 			rRc_BwHalf = new EbwtRangeSource(
-				 ebwtBw_, false, 0xffffffff, false, verbose_, quiet_, 3,  seeded, maqPenalty_, qualOrder_);
+				 ebwtBw_, false, OFF_MASK, false, verbose_, quiet_, 3,  seeded, maqPenalty_, qualOrder_);
 		}
 
 		// Driver wrapper for rRc_Fw
@@ -211,8 +211,8 @@ public:
 		delete drVec;
 
 		// Set up a RangeChaser
-		RangeChaser<String<Dna> > *rchase =
-			new RangeChaser<String<Dna> >(cacheLimit_, cacheFw_, cacheBw_);
+		RangeChaser *rchase =
+			new RangeChaser(cacheLimit_, cacheFw_, cacheBw_);
 
 		return new UnpairedAlignerV2<EbwtRangeSource>(
 			params, dr, rchase,
@@ -221,8 +221,8 @@ public:
 	}
 
 private:
-	Ebwt<String<Dna> >& ebwtFw_;
-	Ebwt<String<Dna> >* ebwtBw_;
+	Ebwt& ebwtFw_;
+	Ebwt* ebwtBw_;
 	bool two_;
 	bool doFw_;
 	bool doRc_;
@@ -232,15 +232,14 @@ private:
 	RangeCache *cacheBw_;
 	const uint32_t cacheLimit_;
 	ChunkPool *pool_;
-	BitPairReference* refs_;
-	vector<String<Dna5> >& os_;
+	BitPairReference *refs_;
+	EList<BTRefString >& os_;
 	const bool maqPenalty_;
 	const bool qualOrder_;
 	const bool strandFix_;
 	const bool rangeMode_;
 	const bool verbose_;
 	const bool quiet_;
-	uint32_t seed_;
 };
 
 /**
@@ -250,12 +249,11 @@ class Paired23mmAlignerV1Factory : public AlignerFactory {
 	typedef RangeSourceDriver<EbwtRangeSource> TRangeSrcDr;
 	typedef ListRangeSourceDriver<EbwtRangeSource> TListRangeSrcDr;
 	typedef CostAwareRangeSourceDriver<EbwtRangeSource> TCostAwareRangeSrcDr;
-	typedef std::vector<TRangeSrcDr*> TRangeSrcDrPtrVec;
+	typedef EList<TRangeSrcDr*> TRangeSrcDrPtrVec;
 public:
 	Paired23mmAlignerV1Factory(
-			Ebwt<String<Dna> >& ebwtFw,
-			Ebwt<String<Dna> >* ebwtBw,
-			bool color,
+			Ebwt& ebwtFw,
+			Ebwt* ebwtBw,
 			bool doFw,
 			bool doRc,
 			bool v1,
@@ -275,7 +273,7 @@ public:
 			uint32_t cacheLimit,
 			ChunkPool *pool,
 			BitPairReference* refs,
-			vector<String<Dna5> >& os,
+			EList<BTRefString >& os,
 			bool reportSe,
 			bool maqPenalty,
 			bool qualOrder,
@@ -286,7 +284,6 @@ public:
 			uint32_t seed) :
 			ebwtFw_(ebwtFw),
 			ebwtBw_(ebwtBw),
-			color_(color),
 			doFw_(doFw),
 			doRc_(doRc),
 			v1_(v1),
@@ -312,8 +309,7 @@ public:
 			strandFix_(strandFix),
 			rangeMode_(rangeMode),
 			verbose_(verbose),
-			quiet_(quiet),
-			seed_(seed)
+			quiet_(quiet)
 	{
 		assert(ebwtBw != NULL);
 		assert(ebwtFw.isInMemory());
@@ -326,16 +322,16 @@ public:
 	virtual Aligner* create() const {
 		HitSinkPerThread* sinkPt = sinkPtFactory_.createMult(2);
 		HitSinkPerThread* sinkPtSe1 = NULL, * sinkPtSe2 = NULL;
-		EbwtSearchParams<String<Dna> >* params =
-			new EbwtSearchParams<String<Dna> >(*sinkPt, os_);
-		EbwtSearchParams<String<Dna> >* paramsSe1 = NULL, * paramsSe2 = NULL;
+		EbwtSearchParams* params =
+			new EbwtSearchParams(*sinkPt, os_);
+		EbwtSearchParams* paramsSe1 = NULL, * paramsSe2 = NULL;
 		if(reportSe_) {
 			sinkPtSe1 = sinkPtFactory_.create();
 			sinkPtSe2 = sinkPtFactory_.create();
 			paramsSe1 =
-				new EbwtSearchParams<String<Dna> >(*sinkPtSe1, os_);
+				new EbwtSearchParams(*sinkPtSe1, os_);
 			paramsSe2 =
-				new EbwtSearchParams<String<Dna> >(*sinkPtSe2, os_);
+				new EbwtSearchParams(*sinkPtSe2, os_);
 		}
 
 		const bool seeded = false;
@@ -361,13 +357,13 @@ public:
 
 		if(do1Fw) {
 			EbwtRangeSource *r1Fw_Bw = new EbwtRangeSource(
-				 ebwtBw_, true, 0xffffffff, true,  verbose_, quiet_, 0, seeded, maqPenalty_, qualOrder_);
+				 ebwtBw_, true, OFF_MASK, true,  verbose_, quiet_, 0, seeded, maqPenalty_, qualOrder_);
 			EbwtRangeSource *r1Fw_Fw = new EbwtRangeSource(
-				&ebwtFw_, true, 0xffffffff, false, verbose_, quiet_, 0, seeded, maqPenalty_, qualOrder_);
+				&ebwtFw_, true, OFF_MASK, false, verbose_, quiet_, 0, seeded, maqPenalty_, qualOrder_);
 			EbwtRangeSource *r1Fw_BwHalf = new EbwtRangeSource(
-				 ebwtBw_, true, 0xffffffff, false, verbose_, quiet_, 2, seeded, maqPenalty_, qualOrder_);
+				 ebwtBw_, true, OFF_MASK, false, verbose_, quiet_, 2, seeded, maqPenalty_, qualOrder_);
 			EbwtRangeSource *r1Fw_FwHalf = two_ ? NULL : new EbwtRangeSource(
-				&ebwtFw_, true, 0xffffffff, false, verbose_, quiet_, 3, seeded, maqPenalty_, qualOrder_);
+				&ebwtFw_, true, OFF_MASK, false, verbose_, quiet_, 3, seeded, maqPenalty_, qualOrder_);
 
 			// Driver wrapper for rFw_Bw
 			EbwtRangeSourceDriver * dr1Fw_Bw = new EbwtRangeSourceDriver(
@@ -426,13 +422,13 @@ public:
 
 		if(do1Rc) {
 			EbwtRangeSource *r1Rc_Fw = new EbwtRangeSource(
-				&ebwtFw_, false, 0xffffffff, true,  verbose_, quiet_, 0, seeded, maqPenalty_, qualOrder_);
+				&ebwtFw_, false, OFF_MASK, true,  verbose_, quiet_, 0, seeded, maqPenalty_, qualOrder_);
 			EbwtRangeSource *r1Rc_Bw = new EbwtRangeSource(
-				 ebwtBw_, false, 0xffffffff, false, verbose_, quiet_, 0, seeded, maqPenalty_, qualOrder_);
+				 ebwtBw_, false, OFF_MASK, false, verbose_, quiet_, 0, seeded, maqPenalty_, qualOrder_);
 			EbwtRangeSource *r1Rc_FwHalf = new EbwtRangeSource(
-				&ebwtFw_, false, 0xffffffff, false, verbose_, quiet_, 2, seeded, maqPenalty_, qualOrder_);
+				&ebwtFw_, false, OFF_MASK, false, verbose_, quiet_, 2, seeded, maqPenalty_, qualOrder_);
 			EbwtRangeSource *r1Rc_BwHalf = two_ ? NULL : new EbwtRangeSource(
-				 ebwtBw_, false, 0xffffffff, false, verbose_, quiet_, 3, seeded, maqPenalty_, qualOrder_);
+				 ebwtBw_, false, OFF_MASK, false, verbose_, quiet_, 3, seeded, maqPenalty_, qualOrder_);
 
 			// Driver wrapper for rRc_Fw
 			EbwtRangeSourceDriver * dr1Rc_Fw = new EbwtRangeSourceDriver(
@@ -491,13 +487,13 @@ public:
 
 		if(do2Fw) {
 			EbwtRangeSource *r2Fw_Bw = new EbwtRangeSource(
-				 ebwtBw_, true, 0xffffffff, true,  verbose_, quiet_, 0, seeded, maqPenalty_, qualOrder_);
+				 ebwtBw_, true, OFF_MASK, true,  verbose_, quiet_, 0, seeded, maqPenalty_, qualOrder_);
 			EbwtRangeSource *r2Fw_Fw = new EbwtRangeSource(
-				&ebwtFw_, true, 0xffffffff, false, verbose_, quiet_, 0, seeded, maqPenalty_, qualOrder_);
+				&ebwtFw_, true, OFF_MASK, false, verbose_, quiet_, 0, seeded, maqPenalty_, qualOrder_);
 			EbwtRangeSource *r2Fw_BwHalf = new EbwtRangeSource(
-				 ebwtBw_, true, 0xffffffff, false, verbose_, quiet_, 2, seeded, maqPenalty_, qualOrder_);
+				 ebwtBw_, true, OFF_MASK, false, verbose_, quiet_, 2, seeded, maqPenalty_, qualOrder_);
 			EbwtRangeSource *r2Fw_FwHalf = two_ ? NULL : new EbwtRangeSource(
-				&ebwtFw_, true, 0xffffffff, false, verbose_, quiet_, 3, seeded, maqPenalty_, qualOrder_);
+				&ebwtFw_, true, OFF_MASK, false, verbose_, quiet_, 3, seeded, maqPenalty_, qualOrder_);
 
 			// Driver wrapper for rFw_Bw
 			EbwtRangeSourceDriver * dr2Fw_Bw = new EbwtRangeSourceDriver(
@@ -555,13 +551,13 @@ public:
 
 		if(do2Rc) {
 			EbwtRangeSource *r2Rc_Fw = new EbwtRangeSource(
-				&ebwtFw_, false, 0xffffffff, true,  verbose_, quiet_, 0, seeded, maqPenalty_, qualOrder_);
+				&ebwtFw_, false, OFF_MASK, true,  verbose_, quiet_, 0, seeded, maqPenalty_, qualOrder_);
 			EbwtRangeSource *r2Rc_Bw = new EbwtRangeSource(
-				 ebwtBw_, false, 0xffffffff, false, verbose_, quiet_, 0, seeded, maqPenalty_, qualOrder_);
+				 ebwtBw_, false, OFF_MASK, false, verbose_, quiet_, 0, seeded, maqPenalty_, qualOrder_);
 			EbwtRangeSource *r2Rc_FwHalf = new EbwtRangeSource(
-				&ebwtFw_, false, 0xffffffff, false, verbose_, quiet_, 2,  seeded, maqPenalty_, qualOrder_);
+				&ebwtFw_, false, OFF_MASK, false, verbose_, quiet_, 2,  seeded, maqPenalty_, qualOrder_);
 			EbwtRangeSource *r2Rc_BwHalf = two_ ? NULL : new EbwtRangeSource(
-				 ebwtBw_, false, 0xffffffff, false, verbose_, quiet_, 3,  seeded, maqPenalty_, qualOrder_);
+				 ebwtBw_, false, OFF_MASK, false, verbose_, quiet_, 3,  seeded, maqPenalty_, qualOrder_);
 
 			// Driver wrapper for rRc_Fw
 			EbwtRangeSourceDriver * dr2Rc_Fw = new EbwtRangeSourceDriver(
@@ -610,16 +606,16 @@ public:
 			}
 		}
 
-		RefAligner<String<Dna5> >* refAligner;
+		RefAligner* refAligner;
 		if(two_) {
-			refAligner = new TwoMMRefAligner<String<Dna5> >(color_, verbose_, quiet_);
+			refAligner = new TwoMMRefAligner(verbose_, quiet_);
 		} else {
-			refAligner = new ThreeMMRefAligner<String<Dna5> >(color_, verbose_, quiet_);
+			refAligner = new ThreeMMRefAligner(verbose_, quiet_);
 		}
 
 		// Set up a RangeChaser
-		RangeChaser<String<Dna> > *rchase =
-			new RangeChaser<String<Dna> >(cacheLimit_, cacheFw_, cacheBw_);
+		RangeChaser* rchase =
+			new RangeChaser(cacheLimit_, cacheFw_, cacheBw_);
 
 		if(v1_) {
 			PairedBWAlignerV1<EbwtRangeSource> *al = new PairedBWAlignerV1<EbwtRangeSource>(
@@ -655,9 +651,8 @@ public:
 	}
 
 private:
-	Ebwt<String<Dna> >& ebwtFw_;
-	Ebwt<String<Dna> >* ebwtBw_;
-	bool color_;
+	Ebwt& ebwtFw_;
+	Ebwt* ebwtBw_;
 	bool doFw_;
 	bool doRc_;
 	bool v1_;
@@ -677,7 +672,7 @@ private:
 	const uint32_t cacheLimit_;
 	ChunkPool *pool_;
 	BitPairReference* refs_;
-	vector<String<Dna5> >& os_;
+	EList<BTRefString >& os_;
 	const bool reportSe_;
 	const bool maqPenalty_;
 	const bool qualOrder_;
@@ -685,7 +680,6 @@ private:
 	const bool rangeMode_;
 	const bool verbose_;
 	const bool quiet_;
-	const uint32_t seed_;
 };
 
 #endif /* ALIGNER_23MM_H_ */
