@@ -2,11 +2,10 @@
 #include <fstream>
 #include <string>
 #include <cassert>
-#include <seqan/index.h>
-#include <seqan/sequence.h>
-#include <seqan/file.h>
 #include <getopt.h>
+
 #include "assert_helpers.h"
+#include "ds.h"
 #include "endian_swap.h"
 #include "ebwt.h"
 #include "formats.h"
@@ -16,6 +15,7 @@
 #include "ref_read.h"
 #include "filebuf.h"
 #include "reference.h"
+#include "sstring.h"
 
 /**
  * \file Driver for the bowtie-build indexing tool.
@@ -25,8 +25,8 @@
 static bool verbose;
 static int sanityCheck;
 static int format;
-static uint32_t bmax;
-static uint32_t bmaxMultSqrt;
+static TIndexOffU bmax;
+static TIndexOffU bmaxMultSqrt;
 static uint32_t bmaxDivN;
 static int dcv;
 static int noDc;
@@ -46,14 +46,16 @@ static bool packed;
 static bool writeRef;
 static bool justRef;
 static int reverseType;
-bool color;
+static int nthreads;
+static string wrapper;
+
 
 static void resetOptions() {
 	verbose      = true;  // be talkative (default)
 	sanityCheck  = 0;     // do slow sanity checks
 	format       = FASTA; // input sequence format
-	bmax         = 0xffffffff; // max blockwise SA bucket size
-	bmaxMultSqrt = 0xffffffff; // same, as multplier of sqrt(n)
+	bmax         = OFF_MASK; // max blockwise SA bucket size
+	bmaxMultSqrt = OFF_MASK; // same, as multplier of sqrt(n)
 	bmaxDivN     = 4;          // same, as divisor of n
 	dcv          = 1024;  // bwise SA difference-cover sample sz
 	noDc         = 0;     // disable difference-cover sample
@@ -62,7 +64,7 @@ static void resetOptions() {
 	showVersion  = 0;     // just print version and quit?
 	doubleEbwt   = true;  // build forward and reverse Ebwts
 	//   Ebwt parameters
-	lineRate     = 6;  // a "line" is 64 bytes
+	lineRate     = Ebwt::default_lineRate;  // a "line" is 64 bytes
 	linesPerSide = 1;  // 1 64-byte line on a side
 	offRate      = 5;  // sample 1 out of 32 SA elts
 	ftabChars    = 10; // 10 chars in initial lookup table
@@ -73,7 +75,8 @@ static void resetOptions() {
 	writeRef     = true;  // write compact reference to .3.ebwt/.4.ebwt
 	justRef      = false; // *just* write compact reference, don't index
 	reverseType  = REF_READ_REVERSE_EACH;
-	color        = false;
+	nthreads     = 1;
+	wrapper.clear();
 }
 
 // Argument constants for getopts
@@ -86,23 +89,36 @@ enum {
 	ARG_PMAP,
 	ARG_NTOA,
 	ARG_USAGE,
-	ARG_NEW_REVERSE
+	ARG_NEW_REVERSE,
+	ARG_THREADS,
+	ARG_WRAPPER
 };
 
 /**
  * Print a detailed usage message to the provided output stream.
  */
 static void printUsage(ostream& out) {
-	out << "Usage: bowtie-build [options]* <reference_in> <ebwt_outfile_base>" << endl
+#ifdef BOWTIE_64BIT_INDEX
+	string tool_name = "bowtie2-build-l";
+#else
+	string tool_name = "bowtie2-build-s";
+#endif
+	if(wrapper == "basic-0") {
+		tool_name = "bowtie-build";
+	}
+
+	out << "Usage: "<< tool_name <<" [options]* <reference_in> <ebwt_outfile_base>" << endl
 	    << "    reference_in            comma-separated list of files with ref sequences" << endl
 	    << "    ebwt_outfile_base       write Ebwt data to files with this dir/basename" << endl
 	    << "Options:" << endl
 	    << "    -f                      reference files are Fasta (default)" << endl
-	    << "    -c                      reference sequences given on cmd line (as <seq_in>)" << endl
-	    << "    -C/--color              build a colorspace index" << endl
-	    << "    -a/--noauto             disable automatic -p/--bmax/--dcv memory-fitting" << endl
+	    << "    -c                      reference sequences given on cmd line (as <seq_in>)" << endl;
+	if(wrapper == "basic-0") {
+		out << "    --large-index           force generated index to be 'large', even if ref" << endl
+		    << "                            has fewer than 4 billion nucleotides" << endl;
+	}
+	out << "    -a/--noauto             disable automatic -p/--bmax/--dcv memory-fitting" << endl
 	    << "    -p/--packed             use packed strings internally; slower, uses less mem" << endl
-	    //<< "    -B                      build both letter- and colorspace indexes" << endl
 	    << "    --bmax <int>            max bucket sz for blockwise suffix-array builder" << endl
 	    //<< "    --bmaxmultsqrt <int>    max bucket sz as multiple of sqrt(ref len)" << endl
 	    << "    --bmaxdivn <int>        max bucket sz as divisor of ref len (default: 4)" << endl
@@ -112,6 +128,7 @@ static void printUsage(ostream& out) {
 	    << "    -3/--justref            just build .3/.4.ebwt (packed reference) portion" << endl
 	    << "    -o/--offrate <int>      SA is sampled every 2^offRate BWT chars (default: 5)" << endl
 	    << "    -t/--ftabchars <int>    # of chars consumed in initial lookup (default: 10)" << endl
+	    << "    --threads <int>         # of threads" << endl
 	    << "    --ntoa                  convert Ns in reference to As" << endl
 	    //<< "    --big --little          endianness (default: little, this host: "
 	    //<< (currentlyBigEndian()? "big":"little") << ")" << endl
@@ -120,8 +137,14 @@ static void printUsage(ostream& out) {
 	    << "    -q/--quiet              verbose output (for debugging)" << endl
 	    << "    -h/--help               print detailed description of tool and its options" << endl
 	    << "    --usage                 print this usage message" << endl
-	    << "    --version               print version information and quit" << endl
-	    ;
+	    << "    --version               print version information and quit" << endl;
+	if(wrapper.empty()) {
+		cerr << endl
+		     << "*** Warning ***" << endl
+		     << "'" << tool_name << "' was run directly.  It is recommended "
+		     << "that you run the wrapper script 'bowtie-build' instead."
+		     << endl << endl;
+	}
 }
 
 static const char *short_options = "qraph?nscfl:i:o:t:h:3C";
@@ -150,8 +173,9 @@ static struct option long_options[] = {
 	{(char*)"ntoa",         no_argument,       0,            ARG_NTOA},
 	{(char*)"justref",      no_argument,       0,            '3'},
 	{(char*)"noref",        no_argument,       0,            'r'},
-	{(char*)"color",        no_argument,       0,            'C'},
+	{(char*)"threads",      required_argument, 0,            ARG_THREADS},
 	{(char*)"usage",        no_argument,       0,            ARG_USAGE},
+	{(char*)"wrapper",      required_argument, 0,            ARG_WRAPPER},
 	{(char*)"new-reverse",  no_argument,       0,            ARG_NEW_REVERSE},
 	{(char*)0, 0, 0, 0} // terminator
 };
@@ -164,7 +188,7 @@ static struct option long_options[] = {
 template<typename T>
 static int parseNumber(T lower, const char *errmsg) {
 	char *endPtr= NULL;
-	T t = (T)strtoll(optarg, &endPtr, 10);
+	T t = (T)strtol(optarg, &endPtr, 10);
 	if (endPtr != NULL) {
 		if (t < lower) {
 			cerr << errmsg << endl;
@@ -185,15 +209,16 @@ static int parseNumber(T lower, const char *errmsg) {
 static void parseOptions(int argc, const char **argv) {
 	int option_index = 0;
 	int next_option;
+	bool bmaxDivNSet = false;
 	do {
 		next_option = getopt_long(
 			argc, const_cast<char**>(argv),
 			short_options, long_options, &option_index);
 		switch (next_option) {
+			case ARG_WRAPPER: wrapper = optarg;	break;
 			case 'f': format = FASTA; break;
 			case 'c': format = CMDLINE; break;
 			case 'p': packed = true; break;
-			case 'C': color = true; break;
 			case 'l':
 				lineRate = parseNumber<int>(3, "-l/--lineRate arg must be at least 3");
 				break;
@@ -219,19 +244,20 @@ static void parseOptions(int argc, const char **argv) {
 				throw 0;
 				break;
 			case ARG_BMAX:
-				bmax = parseNumber<uint32_t>(1, "--bmax arg must be at least 1");
-				bmaxMultSqrt = 0xffffffff; // don't use multSqrt
+				bmax = parseNumber<TIndexOffU>(1, "--bmax arg must be at least 1");
+				bmaxMultSqrt = OFF_MASK; // don't use multSqrt
 				bmaxDivN = 0xffffffff;     // don't use multSqrt
 				break;
 			case ARG_BMAX_MULT:
-				bmaxMultSqrt = parseNumber<uint32_t>(1, "--bmaxmultsqrt arg must be at least 1");
-				bmax = 0xffffffff;     // don't use bmax
+				bmaxMultSqrt = parseNumber<TIndexOffU>(1, "--bmaxmultsqrt arg must be at least 1");
+				bmax = OFF_MASK;     // don't use bmax
 				bmaxDivN = 0xffffffff; // don't use multSqrt
 				break;
 			case ARG_BMAX_DIV:
+				bmaxDivNSet = true;
 				bmaxDivN = parseNumber<uint32_t>(1, "--bmaxdivn arg must be at least 1");
-				bmax = 0xffffffff;         // don't use bmax
-				bmaxMultSqrt = 0xffffffff; // don't use multSqrt
+				bmax = OFF_MASK;         // don't use bmax
+				bmaxMultSqrt = OFF_MASK; // don't use multSqrt
 				break;
 			case ARG_DCV:
 				dcv = parseNumber<int>(3, "--dcv arg must be at least 3");
@@ -240,6 +266,9 @@ static void parseOptions(int argc, const char **argv) {
 				seed = parseNumber<int>(0, "--seed arg must be at least 0");
 				break;
 			case ARG_NTOA: nsToAs = true; break;
+	        	case ARG_THREADS:
+			        nthreads = parseNumber<int>(0, "--threads arg must be at least 1");
+		                break;
 			case ARG_NEW_REVERSE: reverseType = REF_READ_REVERSE; break;
 			case 'a': autoMem = false; break;
 			case 'q': verbose = false; break;
@@ -261,6 +290,9 @@ static void parseOptions(int argc, const char **argv) {
 		     << "extremely slow performance and memory exhaustion.  Perhaps you meant to specify" << endl
 		     << "a small --bmaxdivn?" << endl;
 	}
+	if (!bmaxDivNSet) {
+		bmaxDivN *= nthreads;
+	}
 }
 
 /**
@@ -269,13 +301,13 @@ static void parseOptions(int argc, const char **argv) {
  */
 template<typename TStr>
 static void driver(const string& infile,
-                   vector<string>& infiles,
+                   EList<string>& infiles,
                    const string& outfile,
                    bool reverse = false)
 {
-	vector<FileBuf*> is;
+	EList<FileBuf*> is;
 	bool bisulfite = false;
-	RefReadInParams refparams(color, reverse ? reverseType : REF_READ_FORWARD, nsToAs, bisulfite);
+	RefReadInParams refparams(reverse ? reverseType : REF_READ_FORWARD, nsToAs, bisulfite);
 	assert_gt(infiles.size(), 0);
 	if(format == CMDLINE) {
 		// Adapt sequence strings to stringstreams open for input
@@ -293,12 +325,22 @@ static void driver(const string& infile,
 	} else {
 		// Adapt sequence files to ifstreams
 		for(size_t i = 0; i < infiles.size(); i++) {
-			FILE *f = fopen(infiles[i].c_str(), "r");
-			if (f == NULL) {
-				cerr << "Error: could not open "<< infiles[i] << endl;
-				throw 1;
+			FileBuf *fb = NULL;
+			if (FileBuf::isGzippedFile(infiles[i].c_str())) {
+				gzFile zFp = gzopen(infiles[i].c_str(), "rb");
+				if (zFp == NULL) {
+					std::cerr << "Error: could not open " << infiles[i].c_str() << std::endl;
+					throw 1;
+				}
+				fb = new FileBuf(zFp);
+			} else {
+				FILE *f = fopen(infiles[i].c_str(), "rb");
+				if (f == NULL) {
+					cerr << "Error: could not open "<< infiles[i] << endl;
+					throw 1;
+				}
+				fb = new FileBuf(f);
 			}
-			FileBuf *fb = new FileBuf(f);
 			assert(fb != NULL);
 			assert(!fb->eof());
 			assert(fb->get() == '>');
@@ -310,8 +352,8 @@ static void driver(const string& infile,
 	// Vector for the ordered list of "records" comprising the input
 	// sequences.  A record represents a stretch of unambiguous
 	// characters in one of the input sequences.
-	vector<RefRecord> szs;
-	vector<uint32_t> plens;
+	EList<RefRecord> szs;
+	EList<uint32_t> plens;
 	std::pair<size_t, size_t> sztot;
 	{
 		if(verbose) cout << "Reading reference sizes" << endl;
@@ -319,8 +361,8 @@ static void driver(const string& infile,
 		if(!reverse && (writeRef || justRef)) {
 			// For forward reference, dump it to .3.ebwt and .4.ebwt
 			// files
-			string file3 = outfile + ".3.ebwt";
-			string file4 = outfile + ".4.ebwt";
+			string file3 = outfile + ".3." + gEbwt_ext;
+			string file4 = outfile + ".4." + gEbwt_ext;
 			// Open output stream for the '.3.ebwt' file which will
 			// hold the size records.
 			ofstream fout3(file3.c_str(), ios::binary);
@@ -334,31 +376,11 @@ static void driver(const string& infile,
 			// Read in the sizes of all the unambiguous stretches of
 			// the genome into a vector of RefRecords.  The input
 			// streams are reset once it's done.
-			writeU32(fout3, 1, bigEndian); // endianness sentinel
-			if(color) {
-				refparams.color = false;
-				// Make sure the .3.ebwt and .4.ebwt files contain
-				// nucleotides; not colors
-				int numSeqs = 0;
-				fastaRefReadSizes(is, szs, plens, refparams, &bpout, numSeqs);
-				refparams.color = true;
-				writeU32(fout3, szs.size(), bigEndian); // write # records
-				for(size_t i = 0; i < szs.size(); i++) {
-					szs[i].write(fout3, bigEndian);
-				}
-				szs.clear();
-				plens.clear();
-				// Now read in the colorspace size records; these are
-				// the ones that were indexed
-				int numSeqs2 = 0;
-				sztot = fastaRefReadSizes(is, szs, plens, refparams, NULL, numSeqs2);
-				assert_geq(numSeqs, numSeqs2);
-			} else {
-				int numSeqs = 0;
-				sztot = fastaRefReadSizes(is, szs, plens, refparams, &bpout, numSeqs);
-				writeU32(fout3, szs.size(), bigEndian); // write # records
-				for(size_t i = 0; i < szs.size(); i++) szs[i].write(fout3, bigEndian);
-			}
+			writeU<int32_t>(fout3, 1, bigEndian); // endianness sentinel
+			TIndexOff numSeqs = 0;
+			sztot = fastaRefReadSizes(is, szs, plens, refparams, &bpout, numSeqs);
+			writeU<TIndexOffU>(fout3, (TIndexOffU)szs.size(), bigEndian); // write # records
+			for(size_t i = 0; i < szs.size(); i++) szs[i].write(fout3, bigEndian);
 			if(sztot.first == 0) {
 				cerr << "Error: No unambiguous stretches of characters in the input.  Aborting..." << endl;
 				throw 1;
@@ -371,7 +393,6 @@ static void driver(const string& infile,
 			if(sanityCheck) {
 				BitPairReference bpr(
 					outfile, // ebwt basename
-					color,   // expect color?
 					true,    // sanity check?
 					&infiles,// files to check against
 					NULL,    // sequences to check against
@@ -387,20 +408,8 @@ static void driver(const string& infile,
 		} else {
 			// Read in the sizes of all the unambiguous stretches of the
 			// genome into a vector of RefRecords
-			int numSeqs = 0;
+			TIndexOff numSeqs = 0;
 			sztot = fastaRefReadSizes(is, szs, plens, refparams, NULL, numSeqs);
-#ifndef NDEBUG
-			if(refparams.color) {
-				refparams.color = false;
-				vector<RefRecord> szs2;
-				vector<uint32_t> plens2;
-				int numSeqs2 = 0;
-				fastaRefReadSizes(is, szs2, plens2, refparams, NULL, numSeqs2);
-				assert_leq(numSeqs, numSeqs2);
-				// One less color than base
-				refparams.color = true;
-			}
-#endif
 		}
 	}
 	if(justRef) return;
@@ -408,30 +417,33 @@ static void driver(const string& infile,
 	assert_gt(sztot.second, 0);
 	assert_gt(szs.size(), 0);
 	// Construct Ebwt from input strings and parameters
-	Ebwt<TStr> ebwt(refparams.color ? 1 : 0,
-	                lineRate,
-	                linesPerSide,
-	                offRate,      // suffix-array sampling rate
-	                -1,           // ISA sampling rate
-	                ftabChars,    // number of chars in initial arrow-pair calc
-	                outfile,      // basename for .?.ebwt files
-	                !reverse,     // fw
-	                !entireSA,    // useBlockwise
-	                bmax,         // block size for blockwise SA builder
-	                bmaxMultSqrt, // block size as multiplier of sqrt(len)
-	                bmaxDivN,     // block size as divisor of len
-	                noDc? 0 : dcv,// difference-cover period
-	                is,           // list of input streams
-	                szs,          // list of reference sizes
-	                plens,        // list of not-all-gap reference sequence lengths
-	                sztot.first,  // total size of all unambiguous ref chars
-	                refparams,    // reference read-in parameters
-	                seed,         // pseudo-random number generator seed
-	                -1,           // override offRate
-	                -1,           // override isaRate
-	                verbose,      // be talkative
-	                autoMem,      // pass exceptions up to the toplevel so that we can adjust memory settings automatically
-	                sanityCheck); // verify results and internal consistency
+	Ebwt ebwt(TStr(),
+		  packed,
+		  lineRate,
+		  linesPerSide,
+		  offRate,      // suffix-array sampling rate
+		  -1,           // ISA sampling rate
+		  ftabChars,    // number of chars in initial arrow-pair calc
+		  nthreads,
+		  outfile,      // basename for .?.ebwt files
+		  !reverse,     // fw
+		  !entireSA,    // useBlockwise
+		  bmax,         // block size for blockwise SA builder
+		  bmaxMultSqrt, // block size as multiplier of sqrt(len)
+		  bmaxDivN,     // block size as divisor of len
+		  noDc? 0 : dcv,// difference-cover period
+		  is,           // list of input streams
+		  szs,          // list of reference sizes
+		  plens,        // list of not-all-gap reference sequence lengths
+		  (TIndexOffU)sztot.first,  // total size of all unambiguous ref chars
+		  refparams,    // reference read-in parameters
+		  seed,         // pseudo-random number generator seed
+		  -1,           // override offRate
+		  -1,           // override isaRate
+		  verbose,      // be talkative
+		  autoMem,      // pass exceptions up to the toplevel so that we can adjust memory settings automatically
+		  sanityCheck,  // verify results and internal consistency
+		  false);       // are we building a bt2 index?
 	// Note that the Ebwt is *not* resident in memory at this time.  To
 	// load it into memory, call ebwt.loadIntoMemory()
 	if(verbose) {
@@ -443,30 +455,29 @@ static void driver(const string& infile,
 		// multiple texts, what we'll get back is the joined,
 		// padded string, not a list)
 		ebwt.loadIntoMemory(
-			refparams.color ? 1 : 0,
 			-1,
 			false,
 			false);
-		TStr s2; ebwt.restore(s2);
+		BTRefString s2; ebwt.restore(s2);
 		ebwt.evictFromMemory();
 		{
-			TStr joinedss = Ebwt<TStr>::join(
+			BTRefString joinedss = Ebwt::join<BTRefString >(
 				is,          // list of input streams
 				szs,         // list of reference sizes
-				sztot.first, // total size of all unambiguous ref chars
+				(TIndexOffU)sztot.first, // total size of all unambiguous ref chars
 				refparams,   // reference read-in parameters
 				seed);       // pseudo-random number generator seed
 			if(refparams.reverse == REF_READ_REVERSE) {
-				reverseInPlace(joinedss);
+				joinedss.reverse();
 			}
-			assert_eq(length(joinedss), length(s2));
-			assert_eq(joinedss, s2);
+			assert_eq(joinedss.length(), s2.length());
+			assert(sstr_eq(joinedss, s2));
 		}
 		if(verbose) {
-			if(length(s2) < 1000) {
-				cout << "Passed restore check: " << s2 << endl;
+			if(s2.length() < 1000) {
+				cout << "Passed restore check: " << s2.toZBuf() << endl;
 			} else {
-				cout << "Passed restore check: (" << length(s2) << " chars)" << endl;
+				cout << "Passed restore check: (" << s2.length() << " chars)" << endl;
 			}
 		}
 	}
@@ -478,14 +489,14 @@ extern "C" {
 /**
  * main function.  Parses command-line arguments.
  */
-int bowtie_build(int argc, const char **argv) {
+  int bowtie_build(int argc, const char **argv) {
 	try {
 		// Reset all global state, including getopt state
 		opterr = optind = 1;
 		resetOptions();
 
 		string infile;
-		vector<string> infiles;
+		EList<string> infiles;
 		string outfile;
 
 		parseOptions(argc, argv);
@@ -537,19 +548,19 @@ int bowtie_build(int argc, const char **argv) {
 		// Optionally summarize
 		if(verbose) {
 			cout << "Settings:" << endl
-				 << "  Output files: \"" << outfile << ".*.ebwt\"" << endl
+				 << "  Output files: \"" << outfile << ".*." + gEbwt_ext + "\"" << endl
 				 << "  Line rate: " << lineRate << " (line is " << (1<<lineRate) << " bytes)" << endl
 				 << "  Lines per side: " << linesPerSide << " (side is " << ((1<<lineRate)*linesPerSide) << " bytes)" << endl
 				 << "  Offset rate: " << offRate << " (one in " << (1<<offRate) << ")" << endl
 				 << "  FTable chars: " << ftabChars << endl
 				 << "  Strings: " << (packed? "packed" : "unpacked") << endl
 				 ;
-			if(bmax == 0xffffffff) {
+			if(bmax == OFF_MASK) {
 				cout << "  Max bucket size: default" << endl;
 			} else {
 				cout << "  Max bucket size: " << bmax << endl;
 			}
-			if(bmaxMultSqrt == 0xffffffff) {
+			if(bmaxMultSqrt == OFF_MASK) {
 				cout << "  Max bucket size, sqrt multiplier: default" << endl;
 			} else {
 				cout << "  Max bucket size, sqrt multiplier: " << bmaxMultSqrt << endl;
@@ -581,7 +592,7 @@ int bowtie_build(int argc, const char **argv) {
 			Timer timer(cout, "Total time for call to driver() for forward index: ", verbose);
 			if(!packed) {
 				try {
-					driver<String<Dna, Alloc<> > >(infile, infiles, outfile);
+					driver<BTRefString >(infile, infiles, outfile);
 				} catch(bad_alloc& e) {
 					if(autoMem) {
 						cerr << "Switching to a packed string representation." << endl;
@@ -592,7 +603,7 @@ int bowtie_build(int argc, const char **argv) {
 				}
 			}
 			if(packed) {
-				driver<String<Dna, Packed<Alloc<> > > >(infile, infiles, outfile);
+				driver<S2bDnaString>(infile, infiles, outfile);
 			}
 		}
 		if(doubleEbwt) {
@@ -600,7 +611,7 @@ int bowtie_build(int argc, const char **argv) {
 			Timer timer(cout, "Total time for backward call to driver() for mirror index: ", verbose);
 			if(!packed) {
 				try {
-					driver<String<Dna, Alloc<> > >(infile, infiles, outfile + ".rev", true);
+					driver<BTRefString >(infile, infiles, outfile + ".rev", true);
 				} catch(bad_alloc& e) {
 					if(autoMem) {
 						cerr << "Switching to a packed string representation." << endl;
@@ -611,7 +622,7 @@ int bowtie_build(int argc, const char **argv) {
 				}
 			}
 			if(packed) {
-				driver<String<Dna, Packed<Alloc<> > > >(infile, infiles, outfile + ".rev", true);
+				driver<S2bDnaString>(infile, infiles, outfile + ".rev", true);
 			}
 		}
 		return 0;
@@ -628,5 +639,6 @@ int bowtie_build(int argc, const char **argv) {
 		}
 		return e;
 	}
+  }
 }
-}
+
