@@ -4,20 +4,17 @@
 #include <iostream>
 #include "threading.h"
 
-using tthread::lock_guard;
-
 class SyncLogger {
 public:
-	SyncLogger() {
-	}
+	SyncLogger() : mutex_m() { }
 
 	void msg(const char *s) {
-                lock_guard<MUTEX_T> guard(mutex_m);
+		ThreadSafe ts(&mutex_m);
 		std::cout << s << std::endl;
 	}
 
 	void msg(const std::string& s) {
-                lock_guard<MUTEX_T> guard(mutex_m);
+		ThreadSafe ts(&mutex_m);
 		std::cout << s << std::endl;
 	}
 
