@@ -1,27 +1,26 @@
-#ifndef MULTIKEY_QSORT_H_
+ #ifndef MULTIKEY_QSORT_H_
 #define MULTIKEY_QSORT_H_
 
+#include <algorithm>
 #include <iostream>
-#include <seqan/basic.h>
-#include <seqan/file.h>
-#include <seqan/sequence.h>
-#include "sequence_io.h"
+
 #include "alphabet.h"
 #include "assert_helpers.h"
+#include "btypes.h"
 #include "diff_sample.h"
+#include "ds.h"
+#include "sequence_io.h"
 
 using namespace std;
-using namespace seqan;
 
 /**
  * Swap elements a and b in seqan::String s
  */
 template <typename TStr, typename TPos>
 static inline void swap(TStr& s, size_t slen, TPos a, TPos b) {
-	typedef typename Value<TStr>::Type TAlphabet;
 	assert_lt(a, slen);
 	assert_lt(b, slen);
-	TAlphabet tmp = s[a];
+	char tmp = s[a];
 	s[a] = s[b];
 	s[b] = tmp;
 }
@@ -172,12 +171,12 @@ static inline void vecswap2(TVal* s, size_t slen, TVal* s2, TPos i, TPos j, TPos
 /// Retrieve an int-ized version of the ath character of string s, or,
 /// if a goes off the end of s, return a (user-specified) int greater
 /// than any TAlphabet character - 'hi'.
-#define CHAR_AT(ss, aa) ((length(s[ss]) > aa) ? (int)(Dna)(s[ss][aa]) : hi)
+#define CHAR_AT(ss, aa) ((s[ss].length() > aa) ? (int)(s[ss][aa]) : hi)
 
 /// Retrieve an int-ized version of the ath character of string s, or,
 /// if a goes off the end of s, return a (user-specified) int greater
 /// than any TAlphabet character - 'hi'.
-#define CHAR_AT_SUF(si, off) (((off+s[si]) < hlen) ? ((int)(Dna)(host[off+s[si]])) : (hi))
+#define CHAR_AT_SUF(si, off) (((off+s[si]) < hlen) ? ((int)(host[off+s[si]])) : (hi))
 
 /// Retrieve an int-ized version of the ath character of string s, or,
 /// if a goes off the end of s, return a (user-specified) int greater
@@ -225,16 +224,17 @@ static inline void vecswap2(TVal* s, size_t slen, TVal* s2, TPos i, TPos j, TPos
  * (*prior to* swapping the = regions to the center)
  */
 template<typename THost>
-bool assertPartitionedSuf(const THost& host,
-                          uint32_t *s,
-                          size_t slen,
-                          int hi,
-                          int pivot,
-                          size_t begin,
-                          size_t end,
-                          size_t depth)
+bool assertPartitionedSuf(
+	const THost& host,
+	TIndexOffU *s,
+	size_t slen,
+	int hi,
+	int pivot,
+	size_t begin,
+	size_t end,
+	size_t depth)
 {
-	size_t hlen = length(host);
+	size_t hlen = host.length();
 	int state = 0; // 0 -> 1st = section, 1 -> < section, 2 -> > section, 3 -> 2nd = section
 	for(size_t i = begin; i < end; i++) {
 		switch(state) {
@@ -263,16 +263,17 @@ bool assertPartitionedSuf(const THost& host,
  * (*after* swapping the = regions to the center)
  */
 template<typename THost>
-bool assertPartitionedSuf2(const THost& host,
-                           uint32_t *s,
-                           size_t slen,
-                           int hi,
-                           int pivot,
-                           size_t begin,
-                           size_t end,
-                           size_t depth)
+bool assertPartitionedSuf2(
+	const THost& host,
+	TIndexOffU *s,
+	size_t slen,
+	int hi,
+	int pivot,
+	size_t begin,
+	size_t end,
+	size_t depth)
 {
-	size_t hlen = length(host);
+	size_t hlen = host.length();
 	int state = 0; // 0 -> < section, 1 -> = section, 2 -> > section
 	for(size_t i = begin; i < end; i++) {
 		switch(state) {
@@ -292,53 +293,65 @@ bool assertPartitionedSuf2(const THost& host,
 #endif
 
 /**
- * Assert that the seqan::String s of suffix offsets into seqan::String
- * 'host' is a seemingly legitimate suffix-offset list (at this time,
- * we just check that it doesn't list any suffix twice).
+ * Assert that string s of suffix offsets into string 'host' is a seemingly
+ * legitimate suffix-offset list (at this time, we just check that it doesn't
+ * list any suffix twice).
  */
-static void sanityCheckInputSufs(uint32_t *s, size_t slen) {
+
+static inline void sanityCheckInputSufs(TIndexOffU *s, size_t slen) {
 	assert_gt(slen, 0);
-	for(size_t i = 0; i < slen; i++) {
-		// Actually, it's convenient to allow the caller to provide
-		// suffix offsets thare are off the end of the host string.
-		// See, e.g., build() in diff_sample.cpp.
-		//assert_lt(s[i], length(host));
-		for(size_t j = i+1; j < slen; j++) {
-			assert_neq(s[i], s[j]);
+	// Try keeping the running time <1min
+        if (slen > 4096) {
+		TIndexOffU *s_copy = new TIndexOffU[slen];
+		assert(s_copy != NULL);
+		std::copy(s, s + slen, s_copy);
+		std::sort(s_copy, s_copy + slen);
+		for (size_t i = 0; i < slen - 1; i++)
+			assert_neq(s_copy[i], s_copy[i + 1]);
+		delete[] s_copy;
+        } else {
+		for(size_t i = 0; i < slen; i++) {
+			// Actually, it's convenient to allow the caller to provide
+			// suffix offsets thare are off the end of the host string.
+			// See, e.g., build() in diff_sample.cpp.
+			//assert_lt(s[i], length(host));
+			for(size_t j = i+1; j < slen; j++) {
+				assert_neq(s[i], s[j]);
+			}
 		}
 	}
 }
 
 /**
- * Assert that the seqan::String s of suffix offsets into seqan::String
- * 'host' really are in lexicographical order up to depth 'upto'.
+ * Assert that the string s of suffix offsets into  'host' really are in
+ * lexicographical order up to depth 'upto'.
  */
 template <typename T>
 void sanityCheckOrderedSufs(const T& host,
                             size_t hlen,
-                            uint32_t *s,
+                            TIndexOffU *s,
                             size_t slen,
                             size_t upto,
-                            uint32_t lower = 0,
-                            uint32_t upper = 0xffffffff)
+                            size_t lower = 0,
+                            size_t upper = OFF_MASK)
 {
 	assert_lt(s[0], hlen);
-	upper = min<uint32_t>(upper, slen-1);
+	upper = min<size_t>(upper, slen-1);
 	for(size_t i = lower; i < upper; i++) {
 		// Allow s[i+t] to point off the end of the string; this is
 		// convenient for some callers
 		if(s[i+1] >= hlen) continue;
-		if(upto == 0xffffffff) {
-			assert(dollarLt(suffix(host, s[i]), suffix(host, s[i+1])));
-		} else {
 #ifndef NDEBUG
-			if(prefix(suffix(host, s[i]), upto) > prefix(suffix(host, s[i+1]), upto)) {
+		if(upto == OFF_MASK) {
+			assert(sstr_suf_lt(host, s[i], hlen, host, s[i+1], hlen, false));
+		} else {
+			if(sstr_suf_upto_lt(host, s[i], host, s[i+1], upto, false)) {
 				// operator > treats shorter strings as
 				// lexicographically smaller, but we want to opposite
-				assert(isPrefix(suffix(host, s[i+1]), suffix(host, s[i])));
+				//assert(isPrefix(suffix(host, s[i+1]), suffix(host, s[i])));
 			}
-#endif
 		}
+#endif
 	}
 }
 
@@ -365,15 +378,16 @@ void sanityCheckOrderedSufs(const T& host,
  * keys share a long prefix.
  */
 template<typename T>
-void mkeyQSortSuf(const T& host,
-                  size_t hlen,
-                  uint32_t *s,
-                  size_t slen,
-                  int hi,
-                  size_t begin,
-                  size_t end,
-                  size_t depth,
-                  size_t upto = 0xffffffff)
+void mkeyQSortSuf(
+	const T& host,
+	size_t hlen,
+	TIndexOffU *s,
+	size_t slen,
+	int hi,
+	size_t begin,
+	size_t end,
+	size_t depth,
+	size_t upto = OFF_MASK)
 {
 	// Helper for making the recursive call; sanity-checks arguments to
 	// make sure that the problem actually got smaller.
@@ -454,74 +468,99 @@ void mkeyQSortSuf(const T& host,
  * Toplevel function for multikey quicksort over suffixes.
  */
 template<typename T>
-void mkeyQSortSuf(const T& host,
-                  uint32_t *s,
-                  size_t slen,
-                  int hi,
-                  bool verbose = false,
-                  bool sanityCheck = false,
-                  size_t upto = 0xffffffff)
+void mkeyQSortSuf(
+	const T& host,
+	TIndexOffU *s,
+	size_t slen,
+	int hi,
+	bool verbose = false,
+	bool sanityCheck = false,
+	size_t upto = OFF_MASK)
 {
-	size_t hlen = length(host);
-	assert(!empty(s));
+	size_t hlen = host.length();
+	assert_gt(slen, 0);
 	if(sanityCheck) sanityCheckInputSufs(s, slen);
 	mkeyQSortSuf(host, hlen, s, slen, hi, (size_t)0, slen, (size_t)0, upto);
 	if(sanityCheck) sanityCheckOrderedSufs(host, hlen, s, slen, upto);
 }
 
-/**
- * Just like mkeyQSortSuf but all swaps are applied to s2 as well as s.
- * This is a helpful variant if, for example, the caller would like to
- * see how their input was permuted by the sort routine (in that case,
- * the caller would let s2 be an array s2[] where s2 is the same length
- * as s and s2[i] = i).
- */
+struct QSortRange {
+	size_t begin;
+	size_t end;
+	size_t depth;
+};
+
 template<typename T>
-void mkeyQSortSuf2(const T& host,
-                   size_t hlen,
-                   uint32_t *s,
-                   size_t slen,
-                   uint32_t *s2,
-                   int hi,
-                   size_t begin,
-                   size_t end,
-                   size_t depth,
-                   size_t upto = 0xffffffff)
+void mkeyQSortSuf2(
+	const T& host,
+	size_t hlen,
+	TIndexOffU *s,
+	size_t slen,
+	TIndexOffU *s2,
+	int hi,
+	size_t _begin,
+	size_t _end,
+	size_t _depth,
+	size_t upto = OFF_MASK,
+	EList<size_t>* boundaries = NULL)
 {
-	// Helper for making the recursive call; sanity-checks arguments to
-	// make sure that the problem actually got smaller.
-	#define MQS_RECURSE_SUF_DS(nbegin, nend, ndepth) { \
-		assert(nbegin > begin || nend < end || ndepth > depth); \
-		if(ndepth < upto) { /* don't exceed depth of 'upto' */ \
-			mkeyQSortSuf2(host, hlen, s, slen, s2, hi, nbegin, nend, ndepth, upto); \
-		} \
-	}
-	assert_leq(begin, slen);
-	assert_leq(end, slen);
-	size_t a, b, c, d, /*e,*/ r;
-	size_t n = end - begin;
-	if(n <= 1) return;                 // 1-element list already sorted
-	CHOOSE_AND_SWAP_PIVOT(SWAP2, CHAR_AT_SUF); // pick pivot, swap it into [begin]
-	int v = CHAR_AT_SUF(begin, depth); // v <- randomly-selected pivot value
-	#ifndef NDEBUG
-	{
-		bool stillInBounds = false;
-		for(size_t i = begin; i < end; i++) {
-			if(depth < (hlen-s[i])) {
-				stillInBounds = true;
-				break;
-			} else { /* already fell off this suffix */ }
-		}
-		assert(stillInBounds); // >=1 suffix must still be in bounds
-	}
-	#endif
-	a = b = begin;
-	c = d = /*e =*/ end-1;
+	EList<EList<QSortRange> > block_list;
 	while(true) {
-		// Invariant: everything before a is = pivot, everything
-		// between a and b is <
-		int bc = 0; // shouldn't have to init but gcc on Mac complains
-		while(b <= c && v >= (bc = CHAR_AT_SUF(b, depth))) {
+		size_t begin = 0, end = 0, depth = 0;
+		if(block_list.size() == 0) {
+			begin = _begin;
+			end = _end;
+			depth = _depth;
+		} else {
+			if(block_list.back().size() > 0) {
+				begin = block_list.back()[0].begin;
+				end = block_list.back()[0].end;
+				depth = block_list.back()[0].depth;
+				block_list.back().erase(0);
+			} else {
+				block_list.resize(block_list.size() - 1);
+				if(block_list.size() == 0) {
+					break;
+				}
+			}
+		}
+		if(depth == upto) {
+			if(boundaries != NULL) {
+				(*boundaries).push_back(end);
+			}
+			continue;
+		}
+		assert_leq(begin, slen);
+		assert_leq(end, slen);
+		size_t a, b, c, d, /*e,*/ r;
+		size_t n = end - begin;
+		if(n <= 1) { // 1-element list already sorted
+			if(n == 1 && boundaries != NULL) {
+				boundaries->push_back(end);
+			}
+			continue;
+		}
+		CHOOSE_AND_SWAP_PIVOT(SWAP2, CHAR_AT_SUF); // pick pivot, swap it into [begin]
+		int v = CHAR_AT_SUF(begin, depth); // v <- randomly-selected pivot value
+#ifndef NDEBUG
+		{
+			bool stillInBounds = false;
+			for(size_t i = begin; i < end; i++) {
+				if(depth < (hlen-s[i])) {
+					stillInBounds = true;
+					break;
+				} else { /* already fell off this suffix */ }
+			}
+			assert(stillInBounds); // >=1 suffix must still be in bounds
+		}
+#endif
+		a = b = begin;
+		c = d = /*e =*/ end-1;
+		while(true) {
+			// Invariant: everything before a is = pivot, everything
+			// between a and b is <
+			int bc = 0; // shouldn't have to init but gcc on Mac complains
+			while(b <= c && v >= (bc = CHAR_AT_SUF(b, depth))) {
 			if(v == bc) {
 				SWAP2(s, s2, a, b); a++;
 			}
@@ -538,59 +577,79 @@ void mkeyQSortSuf2(const T& host,
 			c--;
 		}
 		if(b > c) break;
-		SWAP2(s, s2, b, c);
-		b++;
-		c--;
-	}
-	assert(a > begin || c < end-1);                      // there was at least one =s
-	assert_lt(/*e*/d-c, n); // they can't all have been > pivot
-	assert_lt(b-a, n); // they can't all have been < pivot
-	assert(assertPartitionedSuf(host, s, slen, hi, v, begin, end, depth));  // check pre-=-swap invariant
-	r = min(a-begin, b-a); VECSWAP2(s, s2, begin, b-r,   r);  // swap left = to center
-	r = min(d-c, end-d-1); VECSWAP2(s, s2, b,     end-r, r);  // swap right = to center
-	assert(assertPartitionedSuf2(host, s, slen, hi, v, begin, end, depth)); // check post-=-swap invariant
-	r = b-a; // r <- # of <'s
-	if(r > 0) {
-		MQS_RECURSE_SUF_DS(begin, begin + r, depth); // recurse on <'s
-	}
-	// Do not recurse on ='s if the pivot was the off-the-end value;
-	// they're already fully sorted
-	if(v != hi) {
-		MQS_RECURSE_SUF_DS(begin + r, begin + r + (a-begin) + (end-d-1), depth+1); // recurse on ='s
-	}
-	r = d-c;   // r <- # of >'s excluding those exhausted
-	if(r > 0 && v < hi-1) {
-		MQS_RECURSE_SUF_DS(end-r, end, depth); // recurse on >'s
+			SWAP2(s, s2, b, c);
+			b++;
+			c--;
+		}
+		assert(a > begin || c < end-1);                      // there was at least one =s
+		assert_lt(/*e*/d-c, n); // they can't all have been > pivot
+		assert_lt(b-a, n); // they can't all have been < pivot
+		assert(assertPartitionedSuf(host, s, slen, hi, v, begin, end, depth));  // check pre-=-swap invariant
+		r = min(a-begin, b-a); VECSWAP2(s, s2, begin, b-r,   r);  // swap left = to center
+		r = min(d-c, end-d-1); VECSWAP2(s, s2, b,     end-r, r);  // swap right = to center
+		assert(assertPartitionedSuf2(host, s, slen, hi, v, begin, end, depth)); // check post-=-swap invariant
+		r = b-a; // r <- # of <'s
+		EList<QSortRange> tmp1;
+		block_list.push_back(tmp1);
+		block_list.back().clear();
+		if(r > 0) { // recurse on <'s
+			QSortRange tmp2;
+			block_list.back().push_back(tmp2);
+			block_list.back().back().begin = begin;
+			block_list.back().back().end = begin + r;
+			block_list.back().back().depth = depth;
+		}
+		// Do not recurse on ='s if the pivot was the off-the-end value;
+		// they're already fully sorted
+		//if(v != hi) { // recurse on ='s
+			QSortRange tmp3;
+			block_list.back().push_back(tmp3);
+			block_list.back().back().begin = begin + r;
+			block_list.back().back().end = begin + r + (a-begin) + (end-d-1);
+			block_list.back().back().depth = depth + 1;
+			//}
+		r = d-c;   // r <- # of >'s excluding those exhausted
+		if(r > 0 /*&& v < hi-1*/) { // recurse on >'s
+			QSortRange tmp4;
+			block_list.back().push_back(tmp4);
+			block_list.back().back().begin = end - r;
+			block_list.back().back().end = end;
+			block_list.back().back().depth = depth;
+		}
 	}
 }
 
+
 /**
  * Toplevel function for multikey quicksort over suffixes with double
  * swapping.
  */
 template<typename T>
-void mkeyQSortSuf2(const T& host,
-                   uint32_t *s,
-                   size_t slen,
-                   uint32_t *s2,
-                   int hi,
-                   bool verbose = false,
-                   bool sanityCheck = false,
-                   size_t upto = 0xffffffff)
+void mkeyQSortSuf2(
+	const T& host,
+	TIndexOffU *s,
+	size_t slen,
+	TIndexOffU *s2,
+	int hi,
+	bool verbose = false,
+	bool sanityCheck = false,
+	size_t upto = OFF_MASK,
+	EList<size_t>* boundaries = NULL)
 {
-	size_t hlen = length(host);
+	size_t hlen = host.length();
 	if(sanityCheck) sanityCheckInputSufs(s, slen);
-	uint32_t *sOrig = NULL;
+	TIndexOffU *sOrig = NULL;
 	if(sanityCheck) {
-		sOrig = new uint32_t[slen];
-		memcpy(sOrig, s, 4 * slen);
+		sOrig = new TIndexOffU[slen];
+		memcpy(sOrig, s, OFF_SIZE * slen);
 	}
-	mkeyQSortSuf2(host, hlen, s, slen, s2, hi, (size_t)0, slen, (size_t)0, upto);
+	mkeyQSortSuf2(host, hlen, s, slen, s2, hi, (size_t)0, slen, (size_t)0, upto, boundaries);
 	if(sanityCheck) {
 		sanityCheckOrderedSufs(host, hlen, s, slen, upto);
 		for(size_t i = 0; i < slen; i++) {
 			assert_eq(s[i], sOrig[s2[i]]);
 		}
+		delete[] sOrig;
 	}
 }
 
@@ -609,20 +668,24 @@ bool sufDcLt(const T1& host,
              const DifferenceCoverSample<T1>& dc,
              bool sanityCheck = false)
 {
-	uint32_t diff = dc.tieBreakOff(s1, s2);
+	size_t diff = dc.tieBreakOff(s1, s2);
+#ifndef NDEBUG
+	size_t hlen = host.length();
+#endif
 	assert_lt(diff, dc.v());
-	assert_lt(diff, length(host)-s1);
-	assert_lt(diff, length(host)-s2);
+	assert_lt(diff, hlen-s1);
+	assert_lt(diff, hlen-s2);
 	if(sanityCheck) {
-		for(uint32_t i = 0; i < diff; i++) {
+		for(size_t i = 0; i < diff; i++) {
 			assert_eq(host[s1+i], host[s2+i]);
 		}
 	}
 	bool ret = dc.breakTie(s1+diff, s2+diff) < 0;
 #ifndef NDEBUG
-	if(sanityCheck && ret != dollarLt(suffix(host, s1), suffix(host, s2))) {
+	if(sanityCheck && ret != sstr_suf_lt(host, s1, hlen, host, s2, hlen, false)) {
 		assert(false);
 	}
+
 #endif
 	return ret;
 }
@@ -633,7 +696,7 @@ bool sufDcLt(const T1& host,
 template<typename T> inline
 void qsortSufDc(const T& host,
                 size_t hlen,
-                uint32_t* s,
+                TIndexOffU* s,
                 size_t slen,
                 const DifferenceCoverSample<T>& dc,
                 size_t begin,
@@ -656,7 +719,8 @@ void qsortSufDc(const T& host,
 	for(size_t i = begin; i < end-1; i++) {
 		if(sufDcLt(host, s[i], s[end-1], dc, sanityCheck)) {
 			if(sanityCheck)
-				assert(dollarLt(suffix(host, s[i]), suffix(host, s[end-1])));
+				assert(sstr_suf_lt(host, s[i], hlen, host, s[end-1], hlen, false));
+
 			assert_lt(begin + cur, end-1);
 			SWAP(s, i, begin + cur);
 			cur++;
@@ -673,10 +737,10 @@ void qsortSufDc(const T& host,
  * Toplevel function for multikey quicksort over suffixes.
  */
 template<typename T1, typename T2>
-void mkeyQSortSufDcU8(const T1& seqanHost,
+void mkeyQSortSufDcU8(const T1& host1,
                       const T2& host,
                       size_t hlen,
-                      uint32_t* s,
+                      TIndexOffU* s,
                       size_t slen,
                       const DifferenceCoverSample<T1>& dc,
                       int hi,
@@ -684,8 +748,8 @@ void mkeyQSortSufDcU8(const T1& seqanHost,
                       bool sanityCheck = false)
 {
 	if(sanityCheck) sanityCheckInputSufs(s, slen);
-	mkeyQSortSufDcU8(seqanHost, host, hlen, s, slen, dc, hi, 0, slen, 0, sanityCheck);
-	if(sanityCheck) sanityCheckOrderedSufs(seqanHost, hlen, s, slen, 0xffffffff);
+	mkeyQSortSufDcU8(host1, host, hlen, s, slen, dc, hi, 0, slen, 0, sanityCheck);
+	if(sanityCheck) sanityCheckOrderedSufs(host1, hlen, s, slen, OFF_MASK);
 }
 
 /**
@@ -693,32 +757,30 @@ void mkeyQSortSufDcU8(const T1& seqanHost,
  * cover to break the tie.
  */
 template<typename T1, typename T2> inline
-bool sufDcLtU8(const T1& seqanHost,
+bool sufDcLtU8(const T1& host1,
                const T2& host,
                size_t hlen,
-               uint32_t s1,
-               uint32_t s2,
+               size_t s1,
+               size_t s2,
                const DifferenceCoverSample<T1>& dc,
                bool sanityCheck = false)
 {
 	hlen += 0;
-	uint32_t diff = dc.tieBreakOff(s1, s2);
+	size_t diff = dc.tieBreakOff((TIndexOffU)s1, (TIndexOffU)s2);
 	assert_lt(diff, dc.v());
 	assert_lt(diff, hlen-s1);
 	assert_lt(diff, hlen-s2);
 	if(sanityCheck) {
-		for(uint32_t i = 0; i < diff; i++) {
-			assert_eq(host[s1+i], host[s2+i]);
+		for(size_t i = 0; i < diff; i++) {
+			assert_eq(host[s1+i], host1[s2+i]);
 		}
 	}
-	bool ret = dc.breakTie(s1+diff, s2+diff) < 0;
+	bool ret = dc.breakTie((TIndexOffU)(s1+diff), (TIndexOffU)(s2+diff)) < 0;
 	// Sanity-check return value using dollarLt
 #ifndef NDEBUG
-	if(sanityCheck &&
-	   ret != dollarLt(suffix(seqanHost, s1), suffix(seqanHost, s2)))
-	{
-		assert(false);
-	}
+	bool ret2 = sstr_suf_lt(host1, s1, hlen, host, s2, hlen, false);
+	assert(!sanityCheck || ret == ret2);
+
 #endif
 	return ret;
 }
@@ -727,10 +789,10 @@ bool sufDcLtU8(const T1& seqanHost,
  * k log(k)
  */
 template<typename T1, typename T2> inline
-void qsortSufDcU8(const T1& seqanHost,
+void qsortSufDcU8(const T1& host1,
                   const T2& host,
                   size_t hlen,
-                  uint32_t* s,
+                  TIndexOffU* s,
                   size_t slen,
                   const DifferenceCoverSample<T1>& dc,
                   size_t begin,
@@ -751,11 +813,13 @@ void qsortSufDcU8(const T1& seqanHost,
 	SWAP(s, end-1, a); // move pivot to end
 	size_t cur = 0;
 	for(size_t i = begin; i < end-1; i++) {
-		if(sufDcLtU8(seqanHost, host, hlen, s[i], s[end-1], dc, sanityCheck)) {
+		if(sufDcLtU8(host1, host, hlen, s[i], s[end-1], dc, sanityCheck)) {
 #ifndef NDEBUG
-			if(sanityCheck)
-				assert(dollarLt(suffix(seqanHost, s[i]), suffix(seqanHost, s[end-1])));
+			if(sanityCheck) {
+				assert(sstr_suf_lt(host1, s[i], hlen, host1, s[end-1], hlen, false));
+			}
 			assert_lt(begin + cur, end-1);
+
 #endif
 			SWAP(s, i, begin + cur);
 			cur++;
@@ -764,22 +828,19 @@ void qsortSufDcU8(const T1& seqanHost,
 	// Put pivot into place
 	assert_lt(cur, end-begin);
 	SWAP(s, end-1, begin+cur);
-	if(begin+cur > begin) qsortSufDcU8(seqanHost, host, hlen, s, slen, dc, begin, begin+cur);
-	if(end > begin+cur+1) qsortSufDcU8(seqanHost, host, hlen, s, slen, dc, begin+cur+1, end);
+	if(begin+cur > begin) qsortSufDcU8(host1, host, hlen, s, slen, dc, begin, begin+cur);
+	if(end > begin+cur+1) qsortSufDcU8(host1, host, hlen, s, slen, dc, begin+cur+1, end);
 }
 
 #define BUCKET_SORT_CUTOFF (4 * 1024 * 1024)
 #define SELECTION_SORT_CUTOFF 6
 
-// 5 64-element buckets for bucket-sorting A, C, G, T, $
-static uint32_t bkts[4][4 * 1024 * 1024];
-
 /**
  * Straightforwardly obtain a uint8_t-ized version of t[off].  This
  * works fine as long as TStr is not packed.
  */
 template<typename TStr>
-inline uint8_t get_uint8(const TStr& t, uint32_t off) {
+inline uint8_t get_uint8(const TStr& t, size_t off) {
 	return t[off];
 }
 
@@ -789,8 +850,8 @@ inline uint8_t get_uint8(const TStr& t, uint32_t off) {
  * to Dna then to uint8_t.
  */
 template<>
-inline uint8_t get_uint8(const String<Dna, Packed<> >& t, uint32_t off) {
-	return (uint8_t)(Dna)t[off];
+inline uint8_t get_uint8<S2bDnaString>(const S2bDnaString& t, size_t off) {
+	return (uint8_t)t[off];
 }
 
 /**
@@ -799,10 +860,10 @@ inline uint8_t get_uint8(const String<Dna, Packed<> >& t, uint32_t off) {
  */
 template<typename TStr>
 static inline int char_at_suf_u8(const TStr& host,
-                                 uint32_t hlen,
-                                 uint32_t* s,
-                                 uint32_t si,
-                                 uint32_t off,
+                                 size_t hlen,
+                                 TIndexOffU* s,
+                                 size_t si,
+                                 size_t off,
                                  uint8_t hi)
 {
 	return ((off+s[si]) < hlen) ? get_uint8(host, off+s[si]) : (hi);
@@ -810,10 +871,10 @@ static inline int char_at_suf_u8(const TStr& host,
 
 template<typename T1, typename T2>
 static void selectionSortSufDcU8(
-		const T1& seqanHost,
+		const T1& host1,
 		const T2& host,
         size_t hlen,
-        uint32_t* s,
+        TIndexOffU* s,
         size_t slen,
         const DifferenceCoverSample<T1>& dc,
         uint8_t hi,
@@ -823,33 +884,30 @@ static void selectionSortSufDcU8(
         bool sanityCheck = false)
 {
 #define ASSERT_SUF_LT(l, r) \
-	if(sanityCheck && \
-	   !dollarLt(suffix(seqanHost, s[l]), \
-	             suffix(seqanHost, s[r]))) { \
-		cout << "l: " << suffixStr(seqanHost, s[l]) << endl; \
-		cout << "r: " << suffixStr(seqanHost, s[r]) << endl; \
+	if(sanityCheck &&						\
+	   !sstr_suf_lt(host1, s[l], hlen, host1, s[r], hlen, false)) { \
 		assert(false); \
 	}
 
 	assert_gt(end, begin+1);
 	assert_leq(end-begin, SELECTION_SORT_CUTOFF);
 	assert_eq(hi, 4);
-	uint32_t v = dc.v();
+	size_t v = dc.v();
 	if(end == begin+2) {
-		uint32_t off = dc.tieBreakOff(s[begin], s[begin+1]);
+		size_t off = dc.tieBreakOff(s[begin], s[begin+1]);
 		if(off + s[begin] >= hlen ||
 		   off + s[begin+1] >= hlen)
 		{
-			off = 0xffffffff;
+			off = OFF_MASK;
 		}
-		if(off != 0xffffffff) {
+		if(off != OFF_MASK) {
 			if(off < depth) {
-				qsortSufDcU8<T1,T2>(seqanHost, host, hlen, s, slen, dc,
+				qsortSufDcU8<T1,T2>(host1, host, hlen, s, slen, dc,
 				                    begin, end, sanityCheck);
 				// It's helpful for debugging if we call this here
 				if(sanityCheck) {
-					sanityCheckOrderedSufs(seqanHost, hlen, s, slen,
-					                       0xffffffff, begin, end);
+					sanityCheckOrderedSufs(host1, hlen, s, slen,
+					                       OFF_MASK, begin, end);
 				}
 				return;
 			}
@@ -857,15 +915,15 @@ static void selectionSortSufDcU8(
 		}
 	}
 	assert_leq(v, dc.v());
-	uint32_t lim = v;
+	size_t lim = v;
 	assert_geq(lim, 0);
 	for(size_t i = begin; i < end-1; i++) {
-		uint32_t targ = i;
-		uint32_t targoff = depth + s[i];
+		size_t targ = i;
+		size_t targoff = depth + s[i];
 		for(size_t j = i+1; j < end; j++) {
 			assert_neq(j, targ);
-			uint32_t joff = depth + s[j];
-			uint32_t k;
+			size_t joff = depth + s[j];
+			size_t k;
 			for(k = 0; k <= lim; k++) {
 				assert_neq(j, targ);
 				uint8_t jc = (k + joff < hlen)    ? get_uint8(host, k + joff)    : hi;
@@ -909,14 +967,14 @@ static void selectionSortSufDcU8(
 			// difference cover
 			if(k == lim+1) {
 				assert_neq(j, targ);
-				if(sufDcLtU8(seqanHost, host, hlen, s[j], s[targ], dc, sanityCheck)) {
+				if(sufDcLtU8(host1, host, hlen, s[j], s[targ], dc, sanityCheck)) {
 					// j < targ
-					assert(!sufDcLtU8(seqanHost, host, hlen, s[targ], s[j], dc, sanityCheck));
+					assert(!sufDcLtU8(host1, host, hlen, s[targ], s[j], dc, sanityCheck));
 					ASSERT_SUF_LT(j, targ);
 					targ = j;
 					targoff = joff;
 				} else {
-					assert(sufDcLtU8(seqanHost, host, hlen, s[targ], s[j], dc, sanityCheck));
+					assert(sufDcLtU8(host1, host, hlen, s[targ], s[j], dc, sanityCheck));
 					ASSERT_SUF_LT(targ, j); // !
 				}
 			}
@@ -924,7 +982,7 @@ static void selectionSortSufDcU8(
 		if(i != targ) {
 			ASSERT_SUF_LT(targ, i);
 			// swap i and targ
-			uint32_t tmp = s[i];
+			TIndexOffU tmp = s[i];
 			s[i] = s[targ];
 			s[targ] = tmp;
 		}
@@ -933,84 +991,110 @@ static void selectionSortSufDcU8(
 		}
 	}
 	if(sanityCheck) {
-		sanityCheckOrderedSufs(seqanHost, hlen, s, slen,
-		                       0xffffffff, begin, end);
+		sanityCheckOrderedSufs(host1, hlen, s, slen, OFF_MASK, begin, end);
 	}
 }
 
 template<typename T1, typename T2>
 static void bucketSortSufDcU8(
-		const T1& seqanHost,
+		const T1& host1,
 		const T2& host,
         size_t hlen,
-        uint32_t* s,
+        TIndexOffU* s,
         size_t slen,
         const DifferenceCoverSample<T1>& dc,
         uint8_t hi,
-        size_t begin,
-        size_t end,
-        size_t depth,
+        size_t _begin,
+        size_t _end,
+        size_t _depth,
         bool sanityCheck = false)
 {
-	uint32_t cnts[] = { 0, 0, 0, 0, 0 };
-	#define BKT_RECURSE_SUF_DC_U8(nbegin, nend) { \
-		bucketSortSufDcU8<T1,T2>(seqanHost, host, hlen, s, slen, dc, hi, \
-		                         (nbegin), (nend), depth+1, sanityCheck); \
+        // 5 64-element buckets for bucket-sorting A, C, G, T, $
+        TIndexOffU* bkts[4];
+	for(size_t i = 0; i < 4; i++) {
+	        bkts[i] = new TIndexOffU[4 * 1024 * 1024];
 	}
-	assert_gt(end, begin);
-	assert_leq(end-begin, BUCKET_SORT_CUTOFF);
-	assert_eq(hi, 4);
-	if(end == begin+1) return; // 1-element list already sorted
-	if(depth > dc.v()) {
-		// Quicksort the remaining suffixes using difference cover
-		// for constant-time comparisons; this is O(k*log(k)) where
-		// k=(end-begin)
-		qsortSufDcU8<T1,T2>(seqanHost, host, hlen, s, slen, dc, begin, end, sanityCheck);
-		return;
-	}
-	if(end-begin <= SELECTION_SORT_CUTOFF) {
-		// Bucket sort remaining items
-		selectionSortSufDcU8(seqanHost, host, hlen, s, slen, dc, hi,
-		                     begin, end, depth, sanityCheck);
-		if(sanityCheck) {
-			sanityCheckOrderedSufs(seqanHost, hlen, s, slen,
-			                       0xffffffff, begin, end);
-		}
-		return;
-	}
-	for(size_t i = begin; i < end; i++) {
-		uint32_t off = depth+s[i];
-		uint8_t c = (off < hlen) ? get_uint8(host, off) : hi;
-		assert_leq(c, 4);
-		if(c == 0) {
-			s[begin + cnts[0]++] = s[i];
+	EList<EList<size_t> > block_list;
+	bool first = true;
+	while(true) {
+	        size_t begin = 0, end = 0;
+		if(first) {
+		        begin = _begin;
+			end = _end;
+			first = false;
 		} else {
-			bkts[c-1][cnts[c]++] = s[i];
+		        if(block_list.size() == 0) {
+			        break;
+			}
+			if(block_list.back().size() > 1) {
+			        end = block_list.back().back(); 
+				block_list.back().pop_back();
+				begin = block_list.back().back();
+			} else {
+			        block_list.resize(block_list.size() - 1);
+				if(block_list.size() == 0) {
+				        break;
+				}
+			}
+		}
+		size_t depth = block_list.size() + _depth;
+		assert_leq(end-begin, BUCKET_SORT_CUTOFF);
+		assert_eq(hi, 4);
+		if(end <= begin + 1) { // 1-element list already sorted
+		        continue;
+		}
+		if(depth > dc.v()) {
+		        // Quicksort the remaining suffixes using difference cover
+		        // for constant-time comparisons; this is O(k*log(k)) where
+		        // k=(end-begin)
+		        qsortSufDcU8<T1,T2>(host1, host, hlen, s, slen, dc, begin, end, sanityCheck);
+			continue;
+		}
+		if(end-begin <= SELECTION_SORT_CUTOFF) {
+		        // Bucket sort remaining items
+		        selectionSortSufDcU8(host1, host, hlen, s, slen, dc, hi,
+					     begin, end, depth, sanityCheck);
+			if(sanityCheck) {
+			        sanityCheckOrderedSufs(host1, hlen, s, slen,
+						       OFF_MASK, begin, end);
+			}
+			continue;
+		}
+		size_t cnts[] = { 0, 0, 0, 0, 0 };
+		for(size_t i = begin; i < end; i++) {
+		        size_t off = depth + s[i];
+			uint8_t c = (off < hlen) ? get_uint8(host, off) : hi;
+			assert_leq(c, 4);
+			if(c == 0) {
+			        s[begin + cnts[0]++] = s[i];
+			} else {
+			        bkts[c-1][cnts[c]++] = s[i];
+			}
+		}
+		assert_eq(cnts[0] + cnts[1] + cnts[2] + cnts[3] + cnts[4], end - begin);
+		size_t cur = begin + cnts[0];
+		if(cnts[1] > 0) { memcpy(&s[cur], bkts[0], cnts[1] << (OFF_SIZE/4 + 1)); cur += cnts[1]; }
+		if(cnts[2] > 0) { memcpy(&s[cur], bkts[1], cnts[2] << (OFF_SIZE/4 + 1)); cur += cnts[2]; }
+		if(cnts[3] > 0) { memcpy(&s[cur], bkts[2], cnts[3] << (OFF_SIZE/4 + 1)); cur += cnts[3]; }
+		if(cnts[4] > 0) { memcpy(&s[cur], bkts[3], cnts[4] << (OFF_SIZE/4 + 1)); }
+		// This frame is now totally finished with bkts[][], so recursive
+		// callees can safely clobber it; we're not done with cnts[], but
+		// that's local to the stack frame.
+		EList<size_t> tmp;
+		block_list.push_back(tmp);
+		block_list.back().clear();
+		block_list.back().push_back(begin);
+		for(size_t i = 0; i < 4; i++) {
+		        if(cnts[i] > 0) {
+			        block_list.back().push_back(block_list.back().back() + cnts[i]);
+			}
 		}
-	}
-	assert_eq(cnts[0] + cnts[1] + cnts[2] + cnts[3] + cnts[4], end - begin);
-	uint32_t cur = begin + cnts[0];
-	if(cnts[1] > 0) { memcpy(&s[cur], bkts[0], cnts[1] << 2); cur += cnts[1]; }
-	if(cnts[2] > 0) { memcpy(&s[cur], bkts[1], cnts[2] << 2); cur += cnts[2]; }
-	if(cnts[3] > 0) { memcpy(&s[cur], bkts[2], cnts[3] << 2); cur += cnts[3]; }
-	if(cnts[4] > 0) { memcpy(&s[cur], bkts[3], cnts[4] << 2); }
-	// This frame is now totally finished with bkts[][], so recursive
-	// callees can safely clobber it; we're not done with cnts[], but
-	// that's local to the stack frame.
-	cur = begin;
-	if(cnts[0] > 0) {
-		BKT_RECURSE_SUF_DC_U8(cur, cur + cnts[0]); cur += cnts[0];
-	}
-	if(cnts[1] > 0) {
-		BKT_RECURSE_SUF_DC_U8(cur, cur + cnts[1]); cur += cnts[1];
-	}
-	if(cnts[2] > 0) {
-		BKT_RECURSE_SUF_DC_U8(cur, cur + cnts[2]); cur += cnts[2];
-	}
-	if(cnts[3] > 0) {
-		BKT_RECURSE_SUF_DC_U8(cur, cur + cnts[3]);
 	}
 	// Done
+	
+	for(size_t i = 0; i < 4; i++) {
+	        delete [] bkts[i];
+	}
 }
 
 /**
@@ -1033,10 +1117,10 @@ static void bucketSortSufDcU8(
  *     bounds the depth to which the function sorts.
  */
 template<typename T1, typename T2>
-void mkeyQSortSufDcU8(const T1& seqanHost,
+void mkeyQSortSufDcU8(const T1& host1,
                       const T2& host,
                       size_t hlen,
-                      uint32_t* s,
+                      TIndexOffU* s,
                       size_t slen,
                       const DifferenceCoverSample<T1>& dc,
                       int hi,
@@ -1049,7 +1133,7 @@ void mkeyQSortSufDcU8(const T1& seqanHost,
 	// make sure that the problem actually got smaller.
 	#define MQS_RECURSE_SUF_DC_U8(nbegin, nend, ndepth) { \
 		assert(nbegin > begin || nend < end || ndepth > depth); \
-		mkeyQSortSufDcU8(seqanHost, host, hlen, s, slen, dc, hi, nbegin, nend, ndepth, sanityCheck); \
+		mkeyQSortSufDcU8(host1, host, hlen, s, slen, dc, hi, nbegin, nend, ndepth, sanityCheck); \
 	}
 	assert_leq(begin, slen);
 	assert_leq(end, slen);
@@ -1059,18 +1143,18 @@ void mkeyQSortSufDcU8(const T1& seqanHost,
 		// Quicksort the remaining suffixes using difference cover
 		// for constant-time comparisons; this is O(k*log(k)) where
 		// k=(end-begin)
-		qsortSufDcU8<T1,T2>(seqanHost, host, hlen, s, slen, dc, begin, end, sanityCheck);
+		qsortSufDcU8<T1,T2>(host1, host, hlen, s, slen, dc, begin, end, sanityCheck);
 		if(sanityCheck) {
-			sanityCheckOrderedSufs(seqanHost, hlen, s, slen, 0xffffffff, begin, end);
+			sanityCheckOrderedSufs(host1, hlen, s, slen, OFF_MASK, begin, end);
 		}
 		return;
 	}
 	if(n <= BUCKET_SORT_CUTOFF) {
 		// Bucket sort remaining items
-		bucketSortSufDcU8(seqanHost, host, hlen, s, slen, dc,
+		bucketSortSufDcU8(host1, host, hlen, s, slen, dc,
 		                  (uint8_t)hi, begin, end, depth, sanityCheck);
 		if(sanityCheck) {
-			sanityCheckOrderedSufs(seqanHost, hlen, s, slen, 0xffffffff, begin, end);
+			sanityCheckOrderedSufs(host1, hlen, s, slen, OFF_MASK, begin, end);
 		}
 		return;
 	}
@@ -1104,14 +1188,12 @@ void mkeyQSortSufDcU8(const T1& seqanHost,
 		// Invariant: everything after d is = pivot, everything
 		// between c and d is >
 		int cc = 0; // shouldn't have to init but gcc on Mac complains
-		bool hiLatch = true;
+		//bool hiLatch = true;
 		while(b <= c && v <= (cc = CHAR_AT_SUF_U8(c, depth))) {
 			if(v == cc) {
 				SWAP(s, c, d); d--;
 			}
-			else if(hiLatch && cc == hi) {
-
-			}
+			//else if(hiLatch && cc == hi) { }
 			c--;
 		}
 		if(b > c) break;
@@ -1122,10 +1204,8 @@ void mkeyQSortSufDcU8(const T1& seqanHost,
 	assert(a > begin || c < end-1);                      // there was at least one =s
 	assert_lt(d-c, n); // they can't all have been > pivot
 	assert_lt(b-a, n); // they can't all have been < pivot
-	//assert(assertPartitionedSuf(host, s, slen, hi, v, begin, end, depth));  // check pre-=-swap invariant
 	r = min(a-begin, b-a); VECSWAP(s, begin, b-r,   r);  // swap left = to center
 	r = min(d-c, end-d-1); VECSWAP(s, b,     end-r, r);  // swap right = to center
-	//assert(assertPartitionedSuf2(host, s, slen, hi, v, begin, end, depth)); // check post-=-swap invariant
 	r = b-a; // r <- # of <'s
 	if(r > 0) {
 		MQS_RECURSE_SUF_DC_U8(begin, begin + r, depth); // recurse on <'s
